@@ -1,6 +1,6 @@
 /*************************************************************************
  *
- * Copyright (c) 2012 Kohei Yoshida
+ * Copyright (c) 2012-2014 Kohei Yoshida
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -32,6 +32,8 @@
 #include "compat/unique_ptr.hpp"
 #include "global.hpp"
 
+#include <algorithm>
+
 #ifdef MDDS_MULTI_TYPE_VECTOR_USE_DEQUE
 #include <deque>
 #else
@@ -41,6 +43,7 @@
 
 #if defined(MDDS_UNIT_TEST) || defined (MDDS_MULTI_TYPE_VECTOR_DEBUG)
 #include <iostream>
+#include <sstream>
 using std::cout;
 using std::cerr;
 using std::endl;
@@ -116,6 +119,9 @@ protected:
     element_block(size_t n) : base_element_block(_TypeId), m_array(n) {}
     element_block(size_t n, const _Data& val) : base_element_block(_TypeId), m_array(n, val) {}
 
+    template<typename _Iter>
+    element_block(const _Iter& it_begin, const _Iter& it_end) : base_element_block(_TypeId), m_array(it_begin, it_end) {}
+
 public:
     static const element_t block_type = _TypeId;
 
@@ -189,7 +195,11 @@ public:
     {
 #ifdef MDDS_MULTI_TYPE_VECTOR_DEBUG
         if (get_block_type(block) != _TypeId)
-            throw general_error("incorrect block type.");
+        {
+            std::ostringstream os;
+            os << "incorrect block type: expected block type=" << _TypeId << ", passed block type=" << get_block_type(block);
+            throw general_error(os.str());
+        }
 #endif
         return static_cast<_Self&>(block);
     }
@@ -198,7 +208,11 @@ public:
     {
 #ifdef MDDS_MULTI_TYPE_VECTOR_DEBUG
         if (get_block_type(block) != _TypeId)
-            throw general_error("incorrect block type.");
+        {
+            std::ostringstream os;
+            os << "incorrect block type: expected block type=" << _TypeId << ", passed block type=" << get_block_type(block);
+            throw general_error(os.str());
+        }
 #endif
         return static_cast<const _Self&>(block);
     }
@@ -278,7 +292,7 @@ public:
 #ifndef MDDS_MULTI_TYPE_VECTOR_USE_DEQUE
         d.reserve(d.size() + len);
 #endif
-        std::copy(its.first, its.second, std::back_inserter(d));
+        d.insert(d.end(), its.first, its.second);
     }
 
     static void assign_values_from_block(
@@ -330,8 +344,10 @@ public:
         base_element_block& block, size_t pos, const _Iter& it_begin, const _Iter& it_end)
     {
         store_type& d = get(block).m_array;
-        for (_Iter it = it_begin; it != it_end; ++it, ++pos)
-            d[pos] = *it;
+        typename store_type::iterator it_dest = d.begin();
+        std::advance(it_dest, pos);
+        for (_Iter it = it_begin; it != it_end; ++it, ++it_dest)
+            *it_dest = *it;
     }
 
     template<typename _Iter>
@@ -364,6 +380,24 @@ public:
         blk.insert(blk.begin()+pos, it_begin, it_end);
     }
 
+    static size_t capacity(const base_element_block& block)
+    {
+#ifdef MDDS_MULTI_TYPE_VECTOR_USE_DEQUE
+        return 0;
+#else
+        const store_type& blk = get(block).m_array;
+        return blk.capacity();
+#endif
+    }
+
+    static void shrink_to_fit(base_element_block& block)
+    {
+#ifndef MDDS_MULTI_TYPE_VECTOR_USE_DEQUE
+        store_type& blk = get(block).m_array;
+        store_type(blk).swap(blk);
+#endif
+    }
+
 private:
     static std::pair<const_iterator,const_iterator>
     get_iterator_pair(const store_type& array, size_t begin_pos, size_t len)
@@ -386,6 +420,9 @@ protected:
     copyable_element_block(size_t n) : base_type(n) {}
     copyable_element_block(size_t n, const _Data& val) : base_type(n, val) {}
 
+    template<typename _Iter>
+    copyable_element_block(const _Iter& it_begin, const _Iter& it_end) : base_type(it_begin, it_end) {}
+
 public:
     using base_type::get;
 
@@ -404,6 +441,9 @@ protected:
     noncopyable_element_block(size_t n) : base_type(n) {}
     noncopyable_element_block(size_t n, const _Data& val) : base_type(n, val) {}
 
+    template<typename _Iter>
+    noncopyable_element_block(const _Iter& it_begin, const _Iter& it_end) : base_type(it_begin, it_end) {}
+
 public:
     static _Self* clone_block(const base_element_block&)
     {
@@ -437,11 +477,20 @@ struct default_element_block : public copyable_element_block<default_element_blo
     default_element_block(size_t n) : base_type(n) {}
     default_element_block(size_t n, const _Data& val) : base_type(n, val) {}
 
+    template<typename _Iter>
+    default_element_block(const _Iter& it_begin, const _Iter& it_end) : base_type(it_begin, it_end) {}
+
     static self_type* create_block_with_value(size_t init_size, const _Data& val)
     {
         return new self_type(init_size, val);
     }
 
+    template<typename _Iter>
+    static self_type* create_block_with_values(const _Iter& it_begin, const _Iter& it_end)
+    {
+        return new self_type(it_begin, it_end);
+    }
+
     static void overwrite_values(base_element_block&, size_t, size_t)
     {
         // Do nothing.
@@ -474,6 +523,9 @@ struct managed_element_block : public copyable_element_block<managed_element_blo
             m_array.push_back(new _Data(**it));
     }
 
+    template<typename _Iter>
+    managed_element_block(const _Iter& it_begin, const _Iter& it_end) : base_type(it_begin, it_end) {}
+
     ~managed_element_block()
     {
         std::for_each(m_array.begin(), m_array.end(), mdds::default_deleter<_Data>());
@@ -492,6 +544,12 @@ struct managed_element_block : public copyable_element_block<managed_element_blo
         return blk.release();
     }
 
+    template<typename _Iter>
+    static self_type* create_block_with_values(const _Iter& it_begin, const _Iter& it_end)
+    {
+        return new self_type(it_begin, it_end);
+    }
+
     static void overwrite_values(base_element_block& block, size_t pos, size_t len)
     {
         managed_element_block& blk = get(block);
@@ -514,6 +572,9 @@ struct noncopyable_managed_element_block : public noncopyable_element_block<nonc
     noncopyable_managed_element_block() : base_type() {}
     noncopyable_managed_element_block(size_t n) : base_type(n) {}
 
+    template<typename _Iter>
+    noncopyable_managed_element_block(const _Iter& it_begin, const _Iter& it_end) : base_type(it_begin, it_end) {}
+
     ~noncopyable_managed_element_block()
     {
         std::for_each(m_array.begin(), m_array.end(), mdds::default_deleter<_Data>());
@@ -532,6 +593,12 @@ struct noncopyable_managed_element_block : public noncopyable_element_block<nonc
         return blk.release();
     }
 
+    template<typename _Iter>
+    static self_type* create_block_with_values(const _Iter& it_begin, const _Iter& it_end)
+    {
+        return new self_type(it_begin, it_end);
+    }
+
     static void overwrite_values(base_element_block& block, size_t pos, size_t len)
     {
         noncopyable_managed_element_block& blk = get(block);
